@@ -1,5 +1,5 @@
 # 2022.3.20, usage:  import en |  spacy.snts("hello") 
-import json,spacy,os,builtins
+import json,spacy,os,builtins,sqlite3
 from spacy.tokens import DocBin,Doc,Token
 from spacy.tokenizer import Tokenizer
 from spacy.lang.char_classes import ALPHA, ALPHA_LOWER, ALPHA_UPPER, CONCAT_QUOTES, LIST_ELLIPSES, LIST_ICONS
@@ -38,6 +38,21 @@ def spacydoc(snt, use_cache=True):
 	spacydoc.db.conn.commit()
 	return doc 
 
+def cachedoc(snt, table): 
+	'''multiple tables supported,  added 2022.3.27 '''
+	if not hasattr(cachedoc, 'conn'): 
+		cachedoc.conn =	sqlite3.connect("spacy311.sntbs", check_same_thread=False) 
+		cachedoc.conn.execute('PRAGMA synchronous=OFF')
+
+	cachedoc.conn.execute(f'CREATE TABLE IF NOT EXISTS {table} (key varchar(512) PRIMARY KEY, value blob)')
+	item = cachedoc.conn.execute(f'SELECT value FROM "{table}" WHERE key = ? limit 1', (key,)).fetchone()
+	if item	is not None: return spacy.frombs(item[0]) 
+
+	doc = spacy.nlp(snt) 
+	cachedoc.conn.execute(f'REPLACE	INTO {table} (key,	value) VALUES (?,?)',	(snt, spacy.tobs(doc) ))
+	cachedoc.conn.commit()
+	return doc 
+
 if not hasattr(spacy, 'nlp'):
 	from spacy.lang import en
 	spacy.sntbr		= (inst := en.English(), inst.add_pipe("sentencizer"))[0]
@@ -50,16 +65,20 @@ if not hasattr(spacy, 'nlp'):
 	spacy.nps		= lambda doc : {f"{doc[np.end-1].lemma_}/np:{np.text.lower()}" for np in doc.noun_chunks} #book/np:a book
 	spacy.trps		= lambda doc : {f"{t.dep_}_{t.head.pos_}_{t.pos_}:{t.head.lemma_} {t.lemma_}" for t in doc if t.pos_ not in ("PUNCT","PROPN","NUM","SPACE") and t.text.isalpha()} #'dobj_VERB_NOUN:open door';
 	spacy.getdoc	= lambda snt, use_cache=True: spacydoc(snt, use_cache) #lambda snt: ( bs := spacy.txn.get(snt.encode()), doc := spacy.frombs(bs) if bs else spacy.nlp(snt), spacy.txn.put(snt.encode(), spacy.tobs(doc)) if not bs else None )[1] 
+	spacy.redisdoc	= lambda snt, r, prefix="": ( bs := r.get(prefix+snt), doc := spacy.frombs(bs) if bs else spacy.nlp(snt), r.setnx(prefix+snt, spacy.tobs(doc)) if not bs else None )[1] # added 2022.3.29
 	spacy.merge_nps	= spacy.nlp.create_pipe("merge_noun_chunks")
 	spacy.nlp.tokenizer = custom_tokenizer(spacy.nlp)	#nlp.tokenizer.infix_finditer = infix_re.finditer
 	#print([t.text for t in nlp("It's 1.50, up-scaled haven't")])
 	# ['It', "'s", "'", '1.50', "'", ',', 'up-scaled', 'have', "n't"]
 
-postag			= lambda doc:  "_^ " + " ".join([ f"{t.text}_{t.lemma_}_{t.pos_}_{t.tag_}" for t in doc]) + " _$"
 pred_offset		= lambda doc:  (ar := [ t.i for t in doc if t.dep_ == "ROOT"], offset := ar[0] if len(ar) > 0 else 0,  offset/( len(doc) + 0.1) )[-1]
+postag			= lambda doc:  "_^ " + " ".join([ f"{t.text}_{t.lemma_}_{t.pos_}_{t.tag_}" for t in doc]) + " _$"
 non_root_verbs	= lambda doc:  [t for t in doc if t.pos_ == 'VERB' and t.dep_ != 'ROOT'] 
 simple_sent		= lambda doc:  len([t for t in doc if t.pos_ == 'VERB' and t.dep_ != 'ROOT']) <= 0 # else is complex sent 
 compound_snt	= lambda doc:  len([t for t in doc if t.dep_ == 'conj' and t.head.dep_ == 'ROOT']) > 0
+snt_source		= lambda sid, doc: {'type':'snt', 'src': sid, 'snt':doc.text, 'pred_offset': pred_offset(doc), 
+				'postag':'_^ ' + ' '.join([f"{t.text}_{t.lemma_}_{t.pos_}_{t.tag_}" if t.text == t.text.lower() else f"{t.text}_{t.text.lower()}_{t.lemma_}_{t.pos_}_{t.tag_}" for t in doc]) + ' _$',
+			   'tc': len(doc)}
 
 def parse(snt, merge_np= False):
 	''' used in the notebook, for debug '''
@@ -72,6 +91,15 @@ def parse(snt, merge_np= False):
 	'subtree': str([ list(t.subtree) for t in doc]),'children': str([ list(t.children) for t in doc]),
 	}) 
 
+trp_rel		= lambda t:  f"{t.dep_}_{t.head.pos_}_{t.pos_}"  # dobj_VERB_NOUN
+trp_reverse = set({"amod_NOUN_ADJ","nsubj_VERB_NOUN"})
+trp_tok		= lambda doc, arr:  [ t for t in doc if [ t.dep_, t.head.pos_, t.pos_, t.head.lemma_, t.lemma_ ] == arr ] # arr is exactly 5 list 
+gov_dep		= lambda rel, arr : (arr[0], arr[1]) if lemma_order.get(rel, True) else (arr[1], arr[0])  # open door
+hit_trp		= lambda t, _rel, _gov_dep:   _rel == trp_rel(t) and _gov_dep == (t.head.lemma_, t.lemma_)
+trp_high	= lambda doc, i, ihead :   "".join([ f"<b>{t.text_with_ws}</b>" if t.i in (i, ihead) else t.text_with_ws for t in doc ])
+lem_high	= lambda doc, lem :   "".join([ f"<b>{t.text_with_ws}</b>" if t.lemma_ == lem else t.text_with_ws for t in doc ]) # highlight the first lemma 
+
+
 def readline(infile, sepa=None):
 	with open(infile, 'r') as fp:
 		while True:
