@@ -1,11 +1,11 @@
 #!/usr/bin/env python -W ignore::DeprecationWarning 
 # 2022.2.17, upgrade of sntbs, spacy 3.1.1-based, one single file
-import sqlite3,json,collections, fire, traceback,sys
+import sqlite3,json,collections, fire, traceback,sys, time
 from collections import	UserDict,Counter,defaultdict
 from tqdm import tqdm
 
 import en # need 3.1.1 #from en import * #
-from en import terms,verbnet
+from en import terms,verbnet #, snt_source
 attach = lambda doc: ( terms.attach(doc), verbnet.attach(doc), doc.user_data )[-1]  # return ssv, defaultdict(dict)
 
 class Spacybs(UserDict): # change to a neutral name, such as Dbdict ?  then Docbs(Dbdict),   add compress later ? 2021-2-3
@@ -70,6 +70,13 @@ def submit_ssi(dbfile, ssi):
 			conn.execute(f"replace into ssi(s,i) values(?,?)", (f"{k}:{s}",i))
 	conn.commit()
 
+def readline(infile, sepa=None):
+	with open(infile, 'r') as fp:
+		while True:
+			line = fp.readline()
+			if not line: break
+			yield line.strip().split(sepa) if sepa else line.strip()
+
 incr	= lambda si, *names, delta = 1: [si.update({name: delta}) for name in names ] #incr(si, "one", "two")
 doc_fts = lambda conn, rowid, snt, doc: conn.execute(f"insert or ignore into fts(rowid, snt,terms) values(?,?,?)", (rowid, snt, " ".join([ f"{t.pos_}_{t.lemma_}" for t in doc] + [ f"{t.tag_}_{t.lemma_}" for t in doc] + [ f"{t.dep_}_{t.head.pos_}_{t.pos_}_{t.head.lemma_}_{t.lemma_}" for t in doc if t.pos_ not in ('PRON','PUNCT') and t.dep_ in ('dobj','nsubj','advmod','acomp','amod','compound','xcomp','ccomp')]) ))
 
@@ -88,7 +95,27 @@ class util(object):
 		conn.commit()
 		print (f"[fts] finished, {dbfile}")
 
-	def index(self,dbfile, outfile=None, fts=True):  
+	def fts_docsnts(self,spacydbfile, dbfile):  
+		''' doc(rowid, arr['snts']) => fts ,  2022.3.28 '''
+		conn = sqlite3.connect(dbfile, check_same_thread=False)  
+		conn.execute(f'DROP TABLE IF EXISTS fts')
+		conn.execute('''CREATE VIRTUAL TABLE if not exists fts USING fts5(sid, snt, terms, columnsize=0, detail=full,tokenize = "unicode61 remove_diacritics 0 tokenchars '-_'")''') #self.conn.execute('''CREATE VIRTUAL TABLE if not exists fts USING fts5(snt, terms, columnsize=0, detail=none,tokenize = "unicode61 remove_diacritics 0 tokenchars '-_'")''')
+		conn.execute('PRAGMA synchronous=OFF')
+		db = Spacybs(dbfile)
+		for row in conn.execute("select rowid, arr from doc").fetchall(): 
+			try:
+				rowid = row[0]
+				snts = json.loads(row[1]).get('snts',[])
+				for idx, snt in enumerate(snts): 
+					doc = spacy.frombs(db[snt]) if snt in db else spacy.nlp(snt) 
+					kps = " ".join([ f"{t.pos_}_{t.lemma_}" for t in doc] + [ f"{t.tag_}_{t.lemma_}" for t in doc] + [ f"{t.dep_}_{t.head.pos_}_{t.pos_}_{t.head.lemma_}_{t.lemma_}" for t in doc if t.pos_ not in ('PRON','PUNCT') and t.dep_ in ('dobj','nsubj','advmod','acomp','amod','compound','xcomp','ccomp')])
+					conn.execute(f"insert or ignore into fts(sid, snt,terms) values(?,?,?)", (f"{rowid}-{idx}", snt, kps ))
+			except Exception as e:
+				print ("ex:", e, row)
+		conn.commit()
+		print (f"[fts_docsnts] finished, {dbfile}")
+
+	def index(self,dbfile, outfile=None, fts=True, topk=None):  
 		''' clec.spacybs -> clec.sqlite, 2022.2.27 '''
 		if not outfile : outfile  = dbfile.split('.')[0] + ".sqlite" 
 		print (f"[index] started, {dbfile} -> {outfile}", flush=True)
@@ -99,6 +126,7 @@ class util(object):
 				doc = spacy.frombs(bs)
 				terms.zset_ssi(doc, ssi)
 				if fts: doc_fts(conn, rowid, snt, doc) 
+				if topk is not None and rowid > topk : break # add 2022.3.29, wiki.spacybs is too big , 50G memory is not enough
 			except Exception as e:
 				print ("ex:", e, rowid, snt)
 		if fts: conn.commit()
@@ -153,7 +181,7 @@ class util(object):
 		print (f"[si] finished, {dbfile}")
 
 	def train(self, sntfile, dbfile=None): 
-		''' train clec.snt => clec.sntbs, 2021.8.1 '''
+		''' train clec.snt => clec.spacybs, 2021.8.1 '''
 		if not dbfile : dbfile = sntfile.split(".")[0].lower() + ".spacybs"
 		print("started:", sntfile, dbfile,flush=True)
 		db = Spacybs(dbfile)
@@ -167,6 +195,69 @@ class util(object):
 		db.close()
 		print("finished:", sntfile, dbfile)
 
+	def train_json(self, jsonfile, docidx=6, dbfile=None): 
+		''' train spider.json => spider.spacybs, the first column is the docID 2022.3.26 '''
+		if not dbfile : dbfile = jsonfile.split(".")[0].lower() + ".spacybs"
+		print("started:", jsonfile, dbfile,flush=True)
+		start = time.time()
+		db = Spacybs(dbfile)
+		db.conn.execute(f'CREATE TABLE IF NOT EXISTS doc (did int PRIMARY KEY, arr text, snts text) without rowid')
+		for line in readline(jsonfile):
+			try:
+				arr = json.loads(line.strip(), strict=False)
+				did = arr[0]
+				essay = arr[docidx]
+				snts = spacy.snts(essay) 
+				db.conn.execute(f"insert or ignore into doc(did, arr, snts) values(?,?,?)", (did, line, json.dumps(snts)))
+				for snt in snts: 
+					if snt and not snt in db: 
+						db[snt] = spacy.tobs(spacy.nlp(snt))
+				#print (did , time.strftime('%Y.%m.%d %H:%M:%S',time.localtime(time.time()))) #finished: 1k 1k.spacybs 	 using:| 325.5633900165558
+			except Exception as e:
+				print ("parse ex:", e, line)
+		db.close()
+		print("finished:", jsonfile, dbfile, "\t using:|", time.time() - start)
+
+	def addfile(self, infile, dbfile=None): 
+		''' inau.txt => inau.doc-spacybs, doc(arr), spacybs,  2022.3.26 
+		find . -name "*.txt" -exec python spacybs.py addfile {} inau.doc-spacybs \;
+		'''
+		if not dbfile : dbfile = infile.split(".")[0].lower() + ".doc-spacybs"
+		print("started:", infile, dbfile,flush=True)
+		start = time.time()
+		db = Spacybs(dbfile)
+		#db.conn.execute(f'CREATE TABLE IF NOT EXISTS doc (id INTEGER PRIMARY KEY AUTOINCREMENT, arr MEDIUMTEXT, snts MEDIUMTEXT)') # add a md5 ? 
+		db.conn.execute(f'CREATE TABLE IF NOT EXISTS doc (arr MEDIUMTEXT)') # { filename, doc,  snts , 
+		text = open(infile,'r').read().strip()
+		snts = spacy.snts(text) 
+		#db.conn.execute(f"insert or ignore into doc(arr, snts) values(?,?)", (json.dumps({'filename':infile, 'doc':text}), json.dumps(snts)))
+		db.conn.execute(f"insert or ignore into doc(arr) values(?)", (json.dumps({'filename':infile, 'doc':text, 'snts':snts}), ))
+		for snt in snts: 
+			if snt and not snt in db: 
+				db[snt] = spacy.tobs(spacy.nlp(snt))
+		db.close()
+		print("finished:", infile, dbfile, "\t | using:", time.time() - start)
+
+	def load_docjson(self, infile, dbfile=None): 
+		''' biz.json => biz.doc-spacybs, doc(arr), spacybs,  2022.3.26 '''
+		if not dbfile : dbfile = infile.split(".")[0].lower() + ".doc-spacybs"
+		print("started:", infile, dbfile,flush=True)
+		start = time.time()
+		db = Spacybs(dbfile)
+		db.conn.execute(f'drop TABLE IF EXISTS doc')
+		db.conn.execute(f'CREATE TABLE IF NOT EXISTS doc (arr MEDIUMTEXT)') # { filename, doc,  snts , 
+		for line in open(infile,'r').readlines():
+			line = line.strip()
+			if not line: continue
+			db.conn.execute(f"insert or ignore into doc(arr) values(?)", (line, ))
+			arr = json.loads(line) 
+			for snt in arr['snts']: 
+				if snt and not snt in db: 
+					db[snt] = spacy.tobs(spacy.nlp(snt))
+		db.conn.commit()
+		db.close()
+		print("finished:", infile, dbfile, "\t | using:", time.time() - start)
+
 	def topika(self,dbfile, name=None, host='127.0.0.1', port=9311, db_bs=0, db_dm=5):  
 		''' clec.spacybs -> :  snt:clec(list) , {snt}:bs(hash),  2022.1.20 '''
 		import redis
@@ -201,25 +292,25 @@ class util(object):
 					print(">>idsource ex:", ex,"\t|", rowid, snt)
 		print("submit idsource finished:", dbfile, outfile)
 
-	def toes(self,dbfile, idxname=None, eshost='127.0.0.1', esport=9200, batch=100000 , refresh=True):  
-		''' clec.spacybs -> es/clec directly, 2022.2.15
-			{"_id": "140948871-9", "_source": {"rid": "10", "uid": "25110374", "sc": 14, "md5": "da891a7d81f7a5e43b571168cc483b6c dba0b4c99ef37cadfc4bacd61fcefa5b d6b199bfae35246564c598ac78d84c91 38a945eeff5b5a587a26dcc6560e0061 58605af6b50b01f15c0cc3ee2aa75e33 c30566c355ae09ea68673e2940d49d0a 7972c906aef51380310363093e141ef8 dabdd545400415da6d29125bf872"}}
+	def snt_toes(self,dbfile, idxname=None, eshost='127.0.0.1', esport=9200, batch=100000 , refresh=True):  
+		''' clec.spacybs -> es/clec directly, 2022.3.25
+			{"_id": "140948871-9", "_source": {"rid": "10", "uid": "25110374", "sc": 14}}
 		'''
 		from elasticsearch import Elasticsearch,helpers
 		from so import config
+		from en import terms,verbnet, snt_source
 		es	= Elasticsearch([ f"http://{eshost}:{esport}" ])  
 		if not idxname : idxname = dbfile.split('.')[0].lower()
-		if refresh and es.indices.exists(idxname): es.indices.delete(idxname)
-		if not es.indices.exists(idxname): es.indices.create(idxname, config) 
+		if refresh and es.indices.exists(index=idxname): es.indices.delete(index=idxname)
+		if not es.indices.exists(index=idxname): es.indices.create(index=idxname, body=config) 
+
 		print ("toes started:", dbfile, idxname, es, flush=True)
 		actions=[]
 		for rowid, snt, bs in tqdm(Spacybs(dbfile).items()) :
 			try:
 				doc = spacy.frombs(bs) 
-				actions.append({'_op_type':'index', '_index':idxname, '_id': rowid, '_source':{'type':'snt', 'snt':snt,
-				'postag':'^ ' + ' '.join([f"{t.text}_{t.lemma_}_{t.pos_}_{t.tag_}" if t.text == t.text.lower() else f"{t.text}_{t.text.lower()}_{t.lemma_}_{t.pos_}_{t.tag_}" for t in doc]) + ' $',
-				'src': rowid,  'tc': len(doc)}}) #'sid': rowid,
 				ssv = attach(doc) 
+				actions.append({'_op_type':'index', '_index':idxname, '_id': rowid, '_source': snt_source(rowid, doc)}) 
 				for id, source in ssv.items():
 					source.update({"src":rowid}) # sid doesnot work  "sid": rowid, 
 					actions.append({'_op_type':'index', '_index':idxname, '_id': f"{rowid}-{id}", '_source':source})
@@ -234,11 +325,62 @@ class util(object):
 
 		helpers.bulk(client=es,actions=actions, raise_on_error=False)
 		print("toes finished:", dbfile, idxname)
+	
+	def doc_toes(self,dbfile, idxname=None, eshost='127.0.0.1', esport=9200, batch=100000 , refresh=True):  
+		''' inau.doc-spacybs -> es/inau directly, 2022.3.26	'''
+		from elasticsearch import Elasticsearch,helpers
+		from so import config
+		from en import terms,verbnet, snt_source
+		es	= Elasticsearch([ f"http://{eshost}:{esport}" ])  
+		if not idxname : idxname = dbfile.split('.')[0].lower()
+		if refresh and es.indices.exists(index=idxname): es.indices.delete(index=idxname)
+		if not es.indices.exists(index=idxname): es.indices.create(index=idxname, body=config) 
+
+		db = Spacybs(dbfile)
+		print ("doc_toes started:", dbfile, idxname, es, flush=True)
+		actions=[]
+		for row in db.conn.execute("select rowid, arr from doc").fetchall(): 
+			did	 = row[0]
+			arr	 = json.loads(row[1])
+			snts = arr.get('snts',[])
+			arr.update({"did":int(did), "type":"doc"})
+			actions.append({'_op_type':'index', '_index':idxname, '_id': did, '_source': arr}) 
+
+			for idx, snt in enumerate(snts): 
+				try:
+					doc = spacy.frombs(db[snt]) # if snt in db else spacy.nlp(snt) 
+					ssv = attach(doc) 
+					sid = f"{did}-{idx}"
+					actions.append({'_op_type':'index', '_index':idxname, '_id': sid, '_source': snt_source(sid, doc)}) 
+					for id, source in ssv.items():
+						source.update({"src":sid}) # sid doesnot work  "sid": rowid, 
+						actions.append({'_op_type':'index', '_index':idxname, '_id': f"{sid}-{id}", '_source':source})
+					if len(actions) > batch : 
+						helpers.bulk(client=es,actions=actions, raise_on_error=False)
+						actions = []
+						print (did, idx, snt , flush=True) 
+				except Exception as ex:
+					print(">>toes ex:", ex,"\t|", idx, snt)
+					exc_type, exc_value, exc_traceback_obj = sys.exc_info()
+					traceback.print_tb(exc_traceback_obj)
+
+		helpers.bulk(client=es,actions=actions, raise_on_error=False)
+		print("toes finished:", dbfile, idxname)
 
 	def tosnt(self,dbfile):  
 		''' dump snt to console '''
 		for snt in Spacybs(dbfile).keys() :
 			print(snt)
+
+	def merge(self, inpath, attachpath):
+		''' 2022.3.27 '''
+		conn = sqlite3.connect(inpath)
+		conn.text_factory = str
+		cur = conn.cursor()
+		cur.execute(f'attach database {attachpath} as w')
+		cur.execute(f'insert or ignore into spacybs select * from w.spacybs')
+		conn.commit()
+		print ("finished:", inpath, attachpath) 
 			
 if __name__	== '__main__':
 	fire.Fire(util)
@@ -315,4 +457,37 @@ sqlite> select * from ssi where s like '~dobj_VERB_NOUN:difficulty %' order by i
 
 select snt from fts where terms match('dobj_VERB_NOUN_overcome_difficulty')  limit 10;
 select * from ssi where s in ('dobj_VERB_NOUN:overcome difficulty','~amod_NOUN_ADJ:front door')
+
+				actions.append({'_op_type':'index', '_index':idxname, '_id': rowid, '_source':{'type':'snt', 'snt':snt,'pred_offset': en.pred_offset(doc), 
+				'postag':'_^ ' + ' '.join([f"{t.text}_{t.lemma_}_{t.pos_}_{t.tag_}" if t.text == t.text.lower() else f"{t.text}_{t.text.lower()}_{t.lemma_}_{t.pos_}_{t.tag_}" for t in doc]) + ' _$',
+				'src': rowid,  'tc': len(doc)}}) #'sid': rowid,
+
+def spider(self,dbfile, limit="0-10"):  
+		# spider_docs.sqlit3, 2022.3.26
+		print (f"[indexdoc] started, {dbfile}, {limit} ", flush=True)
+		conn = sqlite3.connect(dbfile, check_same_thread=False)  
+		db = Spacybs(dbfile + ".{limit}.spacybs")
+		db.conn.execute(f'CREATE TABLE IF NOT EXISTS snts (did int PRIMARY KEY, snts text) without rowid')
+		for row in conn.execute(f"select id, doc_txt from doc limit {limit.replace('-',',')}").fetchall(): 
+			try:
+				did		= row[0]
+				essay	= row[1]
+				snts	= spacy.snts(essay)
+				db.conn.execute(f"insert or ignore into snts(did, snts) values(?,?)", (did, json.dumps(snts)))
+				for i,snt in enumerate(snts): 
+					doc = spacy.nlp(snt) 
+					if snt and not snt in db: db[snt] = spacy.tobs(spacy.nlp(snt))
+			except Exception as e:
+				print ("ex:", e, row)
+				exc_type, exc_value, exc_traceback_obj = sys.exc_info()
+				traceback.print_tb(exc_traceback_obj)
+		conn.commit()
+		db.close()
+		print (f"[spider] finished, {dbfile}")
+
+ubuntu@es-corpusly-com-105-249:~/pypi/en$ python spacybs.py load_docjson biz.json 
+started: biz.json biz.doc-spacybs
+finished: biz.json biz.doc-spacybs 	 | using: 951.5030801296234
+
+
 '''
\ No newline at end of file
