@@ -1,26 +1,51 @@
 # 2022-2-13  cp from cikuu/bin/es.py 
-import json,fire,sys, redis,os #, spacy
+import json,fire,sys, os, hashlib ,time 
+import warnings
+warnings.filterwarnings("ignore")
+
+import en  
+from en import terms,verbnet
+from en.dims import docs_to_dims
+attach = lambda doc: ( terms.attach(doc), verbnet.attach(doc), doc.user_data )[-1]  # return ssv, defaultdict(dict)
+
+def index_doc(did, doc):  
+	''' arr: additional attr, such as filename , '''
+	arr  = {} #{"did": did}
+	snts = [snt.text for snt in doc.sents]
+	docs = [snt.as_doc() for snt in doc.sents] #spacy.getdoc(snt)
+
+	if len(docs) > 1 : # at least 2 snts will be treated as a document
+		dims = docs_to_dims(snts, docs)
+		dims.update({'type':'doc', "sntnum":len(snts), "wordnum": sum([ len(snt) for snt in snts]), 'tm': time.time()})
+		arr[did] = dims 
+
+	for idx, sdoc in enumerate(docs):
+		arr[f"{did}-{idx}"] = {'type':'snt', 'snt':snts[idx], 'pred_offset': en.pred_offset(sdoc), 
+				'postag':' '.join([f"{t.text}_{t.lemma_}_{t.pos_}_{t.tag_}" if t.text == t.text.lower() else f"{t.text}_{t.text.lower()}_{t.lemma_}_{t.pos_}_{t.tag_}" for t in sdoc]),
+				'src': f"{did}-{idx}",  'tc': len(sdoc)} # src = sentid 
+		ssv = attach(sdoc) 
+		for id, sour in ssv.items():
+			sour.update({"src":f"{did}-{idx}"}) # sid
+			arr[f"{did}-{idx}-{id}"] = sour
+	return arr
+
 from so import * 
-
 class ES(object):
-
 	def __init__(self, host='127.0.0.1',port=9200): 
 		self.es = Elasticsearch([ f"http://{host}:{port}" ])  
 
-	def hello(self): print (self.es)
-
-	def addfolder(self, folder:str, corpus:str, pattern=".txt", idxname:str='docbase'): 
+	def addfolder(self, folder:str, pattern=".txt", idxname=None): 
 		''' folder -> docbase, 2022.1.23 '''
+		if idxname is None : idxname=  folder
 		print("addfolder started:", folder, idxname, self.es, flush=True)
-		if not self.es.indices.exists(idxname): self.es.indices.create(idxname, config)
+		if not self.es.indices.exists(index=idxname): self.es.indices.create(index=idxname, body=config)
 		for root, dirs, files in os.walk(folder):
 			for file in files: 
 				if file.endswith(pattern):
-					self.es.index(index=idxname,id = f"{corpus}-{file}",  body = {"filename":file,'corpus':corpus,'body':open(f"{folder}/{file}",'r', encoding='utf-8').read().strip() })
-					print (file, folder, flush=True)
+					self.add(f"{folder}/{file}", idxname = idxname) 
+					print (f"{folder}/{file}", flush=True)
 		print("addfolder finished:", folder, idxname, self.es, flush=True)
 
-
 	def annotate(self, infile, idxname): 
 		''' 2022.3.24 '''
 		from en import esjson
@@ -35,175 +60,109 @@ class ES(object):
 
 	def add(self, infile, idxname="testdoc"):
 		''' add doc only , 2022.3.25 '''
-		if not self.es.indices.exists(idxname): self.es.indices.create(idxname, config)
+		if not self.es.indices.exists(index=idxname): self.es.indices.create(index=idxname, body=config)
+		start = time.time()
 		text = open(infile, 'r').read().strip() 
 		did	 = hashlib.md5(text.encode("utf8")).hexdigest()
-		res  = self.es.index(index=idxname, body={"doc":text,  "filename": infile, 'type':'doc'}, id = did)
-		print("finished:", infile, idxname,res) 
-
-	def loadjukuu(self, infile, idxname, batch=100000): 
-		''' 2022.1.11 '''
-		print("loadsource started:", infile, idxname, self.es, flush=True)
-		if not self.es.indices.exists(idxname): self.es.indices.create(idxname, config)
-		actions=[]
-		for line in readline(infile):  #{"_id": "140948871-9", "_source": {"rid": "10", "uid": "25110374", "sc": 14, "md5": "da891a7d81f7a5e43b571168cc483b6c dba0b4c99ef37cadfc4bacd61fcefa5b d6b199bfae35246564c598ac78d84c91 38a945eeff5b5a587a26dcc6560e0061 58605af6b50b01f15c0cc3ee2aa75e33 c30566c355ae09ea68673e2940d49d0a 7972c906aef51380310363093e141ef8 dabdd545400415da6d29125bf872"}}
+		self.es.index(index=idxname, body={"doc":text,  "filename": infile, 'type':'doc'}, id = did)
+		ssv  = index_doc(did, spacy.nlp(text))
+		for id, sv in ssv.items(): 
 			try:
-				arr = json.loads(line)
-				arr['content'] = arr['en'] + ' ' + arr['zhseg']
-				del arr['zhseg'] 
-				actions.append({'_op_type':'index', '_index':idxname, "_source": arr})
-				if len(actions) > batch : 
-					helpers.bulk(client=self.es,actions=actions, raise_on_error=False)
-					actions=[]
-					print(arr, flush=True) 
+				self.es.index(index = idxname, id = id, body = sv)
 			except Exception as ex:
-				print(">>callback ex:", ex, line)
-		helpers.bulk(client=self.es,actions=actions, raise_on_error=False)
-		print("loadsource finished:", infile,idxname)
+				print(">>add ex:", ex, id, sv)
+		print(f"{infile} is finished, \t| using: ", time.time() - start) 
 
-	def idsource(self, infile, idxname, batch=100000): 
-		''' {"_id": "140948871-9", "_source": {"rid": "10", "uid": "25110374", "sc": 14, "md5": "da891a7d81f7a5e43b571168cc483b6c dba0b4c99ef37cadfc4bacd61fcefa5b d6b199bfae35246564c598ac78d84c91 38a945eeff5b5a587a26dcc6560e0061 58605af6b50b01f15c0cc3ee2aa75e33 c30566c355ae09ea68673e2940d49d0a 7972c906aef51380310363093e141ef8 dabdd545400415da6d29125bf872"}}
-		added 2022.2.11 '''
-		print("idsource started:", infile, idxname, self.es, flush=True)
-		if not self.es.indices.exists(idxname): self.es.indices.create(idxname, config)
-		actions=[]
-		for line in readline(infile):  
+	def loadsnt(self, infile, idxname=None):
+		''' add doc only , 2022.3.25 '''
+		if idxname is None : idxname = infile.split('.')[0] 
+		if not self.es.indices.exists(index=idxname): self.es.indices.create(index=idxname, body=config)
+		start = time.time()
+		for idx, line in enumerate(open(infile, 'r').readlines()): 
+			ssv  = index_doc(idx, spacy.nlp(line.strip()))
+			for id, sv in ssv.items(): 
+				try:
+					self.es.index(index = idxname, id = id, document = sv) #https://github.com/elastic/elasticsearch-py/issues/1698
+				except Exception as ex:
+					print(">>add ex:", ex, id, sv)
+		print(f"{infile} is finished, \t| using: ", time.time() - start) 
+
+	def sntvec(self, idxname): 
+		''' add snt vec into snt, 2022.3.25 
+		python -m so sntvec testdoc
+		pip install -U sentence-transformers
+		'''
+		from sentence_transformers import SentenceTransformer
+		if not hasattr(fire, 'model'): 
+			fire.model = SentenceTransformer('all-MiniLM-L6-v2')
+			print ("model loaded:", fire.model, flush=True)
+
+		print("sntvec started:", idxname, flush=True) 
+		for doc in helpers.scan(client=self.es, query={"query" : {"match" : {"type":"snt"}} }, index=idxname):
+			sid	= doc['_id']
+			snt	= doc['_source']['snt']
+			vec	= fire.model.encode(snt.strip()).tolist()
+			print (sid, snt, len(vec))
+			self.es.index(index=idxname, body={"_snt":snt,  "sntvec": vec, 'type':'sntvec'}, id = f"{sid}-sntvec")
+		print("sntvec finished:", idxname) 
+
+	def vecso(self, idxname, snt): 
+		''' search with snt vec, 2022.3.27	'''
+		from sentence_transformers import SentenceTransformer
+		if not hasattr(fire, 'model'):  fire.model = SentenceTransformer('all-MiniLM-L6-v2')
+		vec	= fire.model.encode(snt.strip()).tolist()
+		res = self.es.search(index=idxname, body={
+  "query": {
+    "script_score": {
+      "query": {
+        "match_all": {}
+      },
+      "script": {
+        "source": "cosineSimilarity(params.query_vector, 'sntvec') + 1.0",
+        "params": {
+          "query_vector": vec
+        }
+      }
+    }
+  }
+}
+)
+		print (res) 
+
+	def propbank(self, idxname): 
+		'''  add flair semantic tag into snt, 2022.3.25
+		python -m so propbank testdoc
+		'''
+		from flair.models import SequenceTagger
+		from flair.tokenization import SegtokSentenceSplitter
+		from flair.data import Sentence
+		if not hasattr(fire, 'tagger'): 
+			fire.tagger = SequenceTagger.load('frame-fast')  # 115M 
+			print ("flair tagger loaded:", fire.tagger, flush=True)
+
+		print("propbank started:", idxname, flush=True) 
+		for doc in helpers.scan(client=self.es, query={"query" : {"match" : {"type":"snt"}} }, index=idxname):
 			try:
-				arr = json.loads(line)
-				if not '_source' in arr : arr = {"_source": arr} # source only njson 
-				arr.update({'_op_type':'index', '_index':idxname})
-				actions.append(arr)
-				if len(actions) > batch : 
-					helpers.bulk(client=self.es,actions=actions, raise_on_error=False)
-					actions=[]
-					print(arr["_id"], flush=True) 
+				sid	= doc['_id']
+				snt = Sentence(doc['_source']['snt']) #George returned to Berlin to return his hat.
+				fire.tagger.predict(snt)
+				self.es.index(index=idxname, body={"src":sid, 'chunk': snt.to_tagged_string(), 'type':'propbank-snt'}, id = f"{sid}-propbank")
+				for sp in snt.get_spans():  # tag = return.01 
+					self.es.index(index=idxname, body={"src":sid,  "lem": sp.tag.split('.')[0],  "tag": sp.tag, 'lex': sp.text, 'ibeg':sp.start_pos, 'iend': sp.end_pos, 'offset': int(sp.position_string), 'type':'propbank'}, id = f"{sid}-propbank-{sp.position_string}")
 			except Exception as ex:
-				print(">>callback ex:", ex, line)
-		helpers.bulk(client=self.es,actions=actions, raise_on_error=False)
-		print("idsource finished:", infile,idxname)
-
-	def loadsnts(self, infile, idxname, batch=100000): 
-		''' {"eid": 63885890, "ver": "20", "rid": 10, "uid": 6641784, "sc": 28, "snts": ["Robinson was born in an honest, middle - class -- added 2021.12.1 '''
-		print("loadsnts started:", infile, flush=True)
-		if not self.es.indices.exists(idxname): self.es.indices.create(idxname, config)
-		actions=[]
-		for line in readline(infile):  #{"_id": "140948871-9", "_source": {"rid": "10", "uid": "25110374", "sc": 14, "md5": "da891a7d81f7a5e43b571168cc483b6c dba0b4c99ef37cadfc4bacd61fcefa5b d6b199bfae35246564c598ac78d84c91 38a945eeff5b5a587a26dcc6560e0061 58605af6b50b01f15c0cc3ee2aa75e33 c30566c355ae09ea68673e2940d49d0a 7972c906aef51380310363093e141ef8 dabdd545400415da6d29125bf872"}}
-			try:
-				arr = json.loads(line)
-				arr["md5"] = sntmd5( arr['snts'])
-				id = f"{arr['eid']}-{arr['ver']}"
-				del arr['snts']
-				del arr['eid']
-				del arr['ver']
-				actions.append({'_op_type':'index', '_index':idxname,  "_id": id, "_source": arr})
-				if len(actions) > batch : 
-					helpers.bulk(client=self.es,actions=actions, raise_on_error=False)
-					actions=[]
-					print(line) 
-			except Exception as ex:
-				print(">>callback ex:", ex, line)
-		helpers.bulk(client=self.es,actions=actions, raise_on_error=False)
-		print("index_json finished:", infile,idxname)
-
-	def eevjson(self, infile, idxname, batch=100000,  rhost='127.0.0.1', rport=6662, rdb=0): 
-		''' 2021.12.8 '''
-		r = redis.Redis(rhost, port=rport, db=rdb, decode_responses=True)
-		print("load eevjson started:", infile, r, self.es, flush=True)
-		if not self.es.indices.exists(idxname): self.es.indices.create(idxname, config)
-		actions=[]
-		for line in readline(infile):  #{"_id": "140948871-9", "_source": {"rid": "10", "uid": "25110374", "sc": 14, "md5": "da891a7d81f7a5e43b571168cc483b6c dba0b4c99ef37cadfc4bacd61fcefa5b d6b199bfae35246564c598ac78d84c91 38a945eeff5b5a587a26dcc6560e0061 58605af6b50b01f15c0cc3ee2aa75e33 c30566c355ae09ea68673e2940d49d0a 7972c906aef51380310363093e141ef8 dabdd545400415da6d29125bf872"}}
-			try:
-				arr = json.loads(line.strip().replace(', null,',',0,'))
-				eid = int(arr.get('essay_id', 0))
-				ver = str(arr.get('version',0))
-				rid = arr.get('request_id',0)
-				uid = arr.get('user_id',0)
-				ct = arr.get('ctime',0)
-				
-				row = r.get(eid)
-				row = json.loads(row) if row else {}
-				if not ver in row: 
-					print(">> snts missed:", eid, ver, arr['id'], flush=True)
-					row[ver] = spacy.snts(arr['essay'])
-					r.set(eid, json.dumps(row))
-
-				snts = row[ver] 
-				sour = {"rid": int(rid), "uid": int(uid), "sc":len(snts), "md5": sntmd5(snts), "ct": int(ct)}
-				actions.append({'_op_type':'index', '_index':idxname,  "_id": f"{eid}-{ver}", "_source": sour})
-				if len(actions) > batch : 
-					helpers.bulk(client=self.es,actions=actions, raise_on_error=False)
-					actions=[]
-					print(eid, ver, rid, uid, flush=True) 
-			except Exception as ex:
-				print(">>callback ex:", ex, line)
-		helpers.bulk(client=self.es,actions=actions, raise_on_error=False)
-		print("load eevjson finished:", infile,idxname)
-
-	def load_pika_eev(self, ibeg, iend, idxname = 'eevsim', pikahost='192.168.121.3', port=9221, db=2, batch=100000): 
-		''' id: [ibeg, iend) , added 2021.12.2 '''
-		r = redis.Redis(pikahost, port=port, db=db, decode_responses=True)
-		print("index pika started:", ibeg,iend, r)
-		if not self.es.indices.exists(idxname): self.es.indices.create(idxname, config)
-		#{"_id": "140948871-9", "_source": {"rid": "10", "uid": "25110374", "sc": 14, "md5": "da891a7d81f7a5e43b571168cc483b6c dba0b4c99ef37cadfc4bacd61fcefa5b d6b199bfae35246564c598ac78d84c91 38a945eeff5b5a587a26dcc6560e0061 58605af6b50b01f15c0cc3ee2aa75e33 c30566c355ae09ea68673e2940d49d0a 7972c906aef51380310363093e141ef8 dabdd545400415da6d29125bf872"}}
-		for eidv in r.zrangebyscore("eids", ibeg, iend): 
-			try:
-				arr = eidv.strip().split(':') #149304158:15
-				if len(arr) != 2 : continue
-				eid,ver = arr 
-				snts = json.loads(r.hget(eid,'snts'))[ver]
-				self.es.index(index = idxname, id = f"{eid}-{ver}", body = {'rid': int(r.hget(eid,'rid')), 'uid': int(r.hget(eid,'uid')), 'md5': sntmd5(snts), 'sc':len(snts)})
-				print (eidv) 
-			except Exception as ex:
-				print(">>callback ex:", ex, eidv)
-		print("load_pika_eev finished:", ibeg, iend,idxname)
-
+				print ("propbank ex:", ex, doc )
+		print("propbank finished:", idxname) 
+	
 	def init(self, idxname):
 		''' init a new index '''
-		try:
-			if self.es.indices.exists(index=idxname):
-				print("already exists", idxname)
-				self.es.indices.delete(index=idxname)
-			self.es.indices.create(index=idxname, body=config) #, body=snt_mapping
-		except Exception as e:
-			print("exception,", str(e))
+		if self.es.indices.exists(index=idxname):self.es.indices.delete(index=idxname)
+		self.es.indices.create(index=idxname, body=config) #, body=snt_mapping
 		print(">>finished " + idxname )
 
-	def clear(self, idxname):
-		self.es.delete_by_query(idxname, body={"query": {"match_all": {}}})
-
-	def dump(self, idxname):
-		''' --- dump index name to JSON file,ie: pigai06 '''
-		for a in helpers.scan(self.es,query={"query": {"match_all": {}}}, index=idxname):
-			print (a['_id'] + "\t" + json.dumps(a['_source'])) #{'_index': 'pigai06', '_type': '_doc', '_id': '0_40088501_2_0', '_score': None, '_source': {'snt': 'Whether of courses in psychology .', 'postag': 'Whether_IN of_IN courses_NNS in_IN psychology_NN ._.', 'np': ['courses', 'psychology']}, 'sort': [82714]}
-		print(">>finished " , idxname, file=sys.stderr)
-
-	def ids(self, idxname):
-		''' --- dump _id to file,ie: eev '''
-		for a in helpers.scan(self.es,query={"query": {"match_all": {}}}, index=idxname):
-			print(a['_id']) 
-		print(">>finished " , idxname, file=sys.stderr)
-
-	def load(self, kv_json , idxname):
-		''' submit kv_json to es_index, format: _id<TAB><JSON> 20-2-19 '''
-		with open(kv_json,'r') as fp:
-			actions=[]
-			for line in fp.readlines():
-				arr = line.strip().split("\t")
-				if len(arr) >= 2:
-					try:
-						d = json.loads(arr[1])
-						actions.append({'_op_type':'index', '_index':idxname, '_id':arr[0].strip() , '_source':d})
-					except Exception as e:
-						print("ex:", e, line)
-			helpers.bulk(client=es,actions=actions, raise_on_error=False)
-		print(">>finished " , kv_json, idxname )
-
-	def keys(self, idxname):
-		''' --- dump index name,ie: pigai06 '''
-		query={"query" : {"match_all" : {}}}
-		scanResp= helpers.scan(client= es, query=query, scroll= "10m", index= idxname , timeout="10m")
-		for resp in scanResp:
-			print(resp['_id'])
+	def clear(self,idxname): self.es.delete_by_query(index=idxname, body={"query": {"match_all": {}}})
+	def dump(self, idxname): [print (doc['_id'] + "\t" + json.dumps(doc['_source']))  for doc in helpers.scan(self.es,query={"query": {"match_all": {}}}, index=idxname)]
+	def keys(self, idxname): [print(resp['_id']) for resp in  helpers.scan(client=self.es, query={"query" : {"match_all" : {}}}, scroll= "10m", index= idxname , timeout="10m") ]
+	def hello(self): print (self.es)
 
 if __name__ == '__main__':
 	fire.Fire(ES)
@@ -413,5 +372,8 @@ ubuntu@ubuntu:/ftp/esmd5$ nohup python3 -m cikuu.bin.es loadjson eev2021.esmd5 e
 					}
 				  ]
 				},
+			
+		print(">>finished " , idxname, file=sys.stderr)
 
+self.es.index(index=idxname,id = f"{corpus}-{file}",  body = {"filename":file,'corpus':corpus,'body':open(f"{folder}/{file}",'r', encoding='utf-8').read().strip() })
 '''
\ No newline at end of file
