@@ -26,6 +26,17 @@ def readline(infile, sepa=None):
 			if not line: break
 			yield line.strip().split(sepa) if sepa else line.strip()
 
+def process(infile, outfile, asjson=True, func = lambda x: x):
+	''' line processor, added 2022.3.20  '''
+	print ("process started:", infile, outfile, flush=True)
+	with open(outfile, 'w') as fw: 
+		for line in readline(infile):
+			try:
+				fw.write( func( json.loads(line.strip(), strict=False) if asjson else line.strip())  + "\n")
+			except Exception as ex:
+				print ("process ex:", ex, line) 
+	print ('process finished:', infile) 
+
 def hset_if_greater(r, key, eid, ver ): 
 	res = r.hget(key, eid)
 	try: 
